#include "config.h"
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <sys/stat.h>
#include <time.h>
#include <unistd.h>
#include "itoa.h"
#include "defines.h"
#include "mystring.h"
#include "fdbuf.h"

#define fail(MSG) do{ fout.write("nullmailer-queue: "); fout.writeln(MSG); return false; }while(0)

bool is_dir(const char* path)
{
  struct stat buf;
  return !stat(path, &buf) && S_ISDIR(buf.st_mode);
}

bool is_exist(const char* path)
{
  struct stat buf;
  return !stat(path, &buf);
}

int fsyncdir(const char* path)
{
  int fd = open(path, O_RDONLY);
  if(fd == -1)
    return 0;
  int result = fsync(fd);
  if(result == -1 && errno != EIO)
    result = 0;
  close(fd);
  return result;
}

void trigger()
{
  int fd = open(QUEUE_TRIGGER, O_WRONLY|O_NONBLOCK, 0666);
  if(fd == -1)
    return;
  char x = 0;
  write(fd, &x, 1);
  close(fd);
}

bool validate_addr(const mystring& addr)
{
  int i = addr.find_last('@');
  if(i < 0)
    return false;
  if(addr.find('.', i) < 0)
    return false;
  return true;
}

bool copyenv(fdobuf& out)
{
  mystring str;
  if(!fin.getline(str) || !str)
    fail("Could not read envelope sender.");
  if(!validate_addr(str))
    fail("Envelope sender address is invalid.");
  if(!out.writeln(str))
    fail("Could not write envelope sender.");
  unsigned count=0;
  while(fin.getline(str) && !!str) {
    if(!validate_addr(str))
      fail("Envelope recipient address is invalid.");
    if(!out.writeln(str))
      fail("Could not write envelope recipient.");
    ++count;
  }
  if(count == 0)
    fail("No envelope recipients read.");
  if(!out.write("\n"))
    fail("Could not write extra blank line to destination.");
  return true;
}

bool makereceived(fdobuf& out)
{
  mystring line("Received: (nullmailer pid ");
  line += itoa(getpid());
  line += " invoked by uid ";
  line += itoa(getuid());
  line += ");\n\t";
  time_t t;
  t = time(0);
  char buf[100];
  if(!strftime(buf, 100, "%a, %d %b %Y %H:%M:%S -0000\n", gmtime(&t)))
    fail("Error generating a date string.");
  line += buf;
  if(!out.write(line))
    fail("Could not write received line to message.");
  return true;
}

bool dump(int fd)
{
  fdobuf out(fd);
  if(!copyenv(out))
    return false;
  if(!makereceived(out))
    return false;
  if(!fdbuf_copy(fin, out))
    fail("Error copying the message to the queue file.");
  if(!out.flush())
    fail("Error flushing the output file.");
  if(!out.close())
    fail("Error closing the output file.");
  return true;
}

bool deliver()
{
  if(!is_dir(QUEUE_MSG_DIR) || !is_dir(QUEUE_TMP_DIR))
    fail("Installation error: queue directory is invalid.");

  // Notes:
  // - temporary file name is unique to the currently running
  //   nullmailer-queue program
  // - destination file name is unique to the system
  // - if the temporary file previously existed, it did so because
  //   the previous nullmailer-queue process crashed, and it can be
  //   safely overwritten
  pid_t pid = getpid();
  const mystring unique(itoa(time(0)), ".", itoa(pid));
  const mystring tmpfile(QUEUE_TMP_DIR, itoa(pid));
  const mystring newfile = mystring(QUEUE_MSG_DIR, itoa(time(0)),
				    ".", itoa(pid));

  int out = open(tmpfile.c_str(), O_CREAT|O_WRONLY|O_TRUNC, 0666);
  if(out < 0)
    fail("Could not open temporary file for writing");
  if(!dump(out)) {
    unlink(tmpfile.c_str());
    return false;
  }
  if(link(tmpfile.c_str(), newfile.c_str()))
    fail("Error linking the temp file to the new file.");
  if(fsyncdir(QUEUE_MSG_DIR))
    fail("Error syncing the new directory.");
  if(unlink(tmpfile.c_str()))
    fail("Error unlinking the temp file.");
  return true;
}

int main(int, char*[])
{
  if(!deliver())
    return 1;
  trigger();
  return 0;
}
